#!encoding=utf-8
import time
import threading

"""
    Cache().setex(k, v, timeout)
    Cache().get(k)
"""

HB = 10     # 检查频率 秒

class Cache(object):
    m = {}

    def __new__(cls, *args, **kw):
        if not hasattr(cls, '_instance'):
            orig = super(Cache, cls)
            cls._instance = orig.__new__(cls, *args, **kw)

            t = threading.Thread(target=cls._instance.timer)
            t.setDaemon(True)
            t.start()

        return cls._instance

    def setex(self, k, v, ex = None):
        """
            k, v, ex = expire
        """
        if ex: ex = time.time() + ex
        self.m[k] = (v, ex)

    def get(self, k):
        return self.m.get(k)[0] if self.m.has_key(k) else None

    def timer(self):
        while True:
            now = time.time()
            for k, v in self.m.items():
                if v[1] and now > v[1]:
                    self.m.pop(k, None)

            time.sleep(HB)

if __name__ == '__main__':
    Cache().setex(1, 2, 6)
    for i in range(10):
        print(Cache().get(1))
        time.sleep(1)

    import tornado.ioloop
    tornado.ioloop.IOLoop.instance().start()






# -*- coding:utf-8 -*-
"""
    author comger@gmail.com
    Consisten hash 
"""
from hashlib import md5
from bisect import bisect_right

class ConsistentHash(object):
    """
    算法思路：
    1. 在N个机器中、每台M个节点、N*M 个节点形成节点环
    2. 计算每个机器拥有的节点Node
    3. 新内容key添加时,get_node(key)获取key被分配的node;及get_host(key)获取key 被分配到的机器
    
    * 节点的引入：保证每台机器负载均衡
    """
    def __init__(self, hosts, replicas = 10):
        self._hosts = {}
        self._ring = []
        self._length = len(hosts)*replicas
        self._build(hosts, replicas)
    
    def _build(self, hosts, replicas):
        for host in hosts:
            for i in xrange(replicas):
                key = "{0}_{1}".format(host,i)
                hsh = self._hash(key)
                
                self._hosts[str(hsh)] = host
                print(bisect_right(self._ring, hsh),hsh,host)
                self._ring.insert(bisect_right(self._ring, hsh), hsh)
                print(self._ring)


    def _hash(self,s):
        return hash(md5(s).digest()) % self._length

    
    def get_node(self, key):
        hsh = self._hash(key)
        index = bisect_right(self._ring, hsh)
        if index >= len(self._ring): index = 0

        return self._ring[index]

    def get_host(self, key):
        return self._hosts[str(self.get_node(key))]

    
    
if __name__ == '__main__':
    from random import sample
    from string import letters

    '''
    loop = 100000
    hosts = ["192.168.1.%d" % i for i in xrange(1, 10)]
    ch = ConsistentHash(hosts,replicas=100)
    
    rnd_key = lambda: "".join(sample(letters, 10))
    count = {}

    for i in xrange(loop):
        host = ch.get_host(rnd_key())
        count[host] = count[host] + 1 if host in count else 1

    avg = loop / len(hosts)

    for h in sorted(count.iterkeys()):
        c = count[h]
        print("{0:15} {1:8} {2:8.2f}%".format(h, c, float(c) / avg * 100))

        if c< avg*0.6:
            print("ERROR", h,c)
    '''
    
    dh = 'asdfasd'
    hosts = ["192.168.1.%d" % i for i in xrange(1, 5)]
    ch = ConsistentHash(hosts,replicas=10)
    print(ch.get_host(dh))




#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
    author comger@gmail.com
"""
import sys
import json
import tornado.options

from datetime import datetime
from pprint import pprint

from tornado.options import define, options
from tornado.httpserver import HTTPServer
from tornadorpc.json import JSONRPCHandler

from kpages.router import load_handlers
from kpages.utility import refresh_config,set_default_encoding


define("rpcport", default=None, help="run on the given port", type=int)
'''
define("config", default='setting.py', help="set config for server")
define("ip",  help="bind accept ip  for server")
define("debug", default=None,  help="Debug Mode")
define("ndebug",  help="No Debug Mode")
'''

def load_rpc_handlers():
    member_filter = lambda m : isinstance(m, JSONRPCHandler) and hasattr(m, '__urls__') and m.__urls__
    return load_handlers(__conf__.RPC_DIR)


def start_server():
    settings = {"debug": __conf__.DEBUG}
    handlers = load_rpc_handlers()
    callback(handlers)
    app = tornado.web.Application(handlers, **settings)


    if __conf__.DEBUG:
        app.listen(__conf__.PORT, address=__conf__.BIND_IP, max_buffer_size = __conf__.max_buffer_size)
    else:
        server = HTTPServer(app, xheaders=True, max_buffer_size = __conf__.max_buffer_size)
        server.bind(__conf__.PORT, address= __conf__.BIND_IP )
        server.start(0)

    loop_instance = tornado.ioloop.IOLoop.instance()
    for (route, handler) in handlers:
        try:
            setattr(handler, '_server', loop_instance)
        except AttributeError:
            handler._server = loop_instance
    
    loop_instance.start()
    return loop_instance
        



def run():
    set_default_encoding()
    tornado.options.parse_command_line()
    opts = options
    refresh_config(opts.config)

    __conf__.PORT = opts.rpcport or __conf__.RPC_PORT
    __conf__.DEBUG = opts.debug or __conf__.DEBUG
    __conf__.BIND_IP = opts.ip or __conf__.BIND_IP
    
    start_server()
    

def callback(handlers):
    settings = __conf__.__dict__
    print("Start time: {0}".format(datetime.now().isoformat(" ")))
    print("Config Params")
    for k in sorted(settings.keys()):
        if k.startswith("__"):
            continue
        print("  {0:<40} : {1}".format(k, settings[k]))
    

    print("RPC Router Handlers")
    for h in handlers:
        print('  {0:<50} : {1}'.format(h[1],h[0]))

if __name__ == "__main__":
    try:
        run()
    except KeyboardInterrupt:
        print('exit server ')

# vim: ts=4 sw=4 sts=4 expandtab



# -*- coding:utf-8 -*-

"""
    Reids MQ Service



    作者: Q.yuhen, comger@gmail.com
    创建: 2011-08-01

    历史:
        2011-08-03  + srvcmd 装饰器
        2011-08-04  * 取消 Service callback，改为自动查找 service function。
                    + 增加 _send_pack, _unpack。
        2011-08-07  * 将 PSConsumer 从 Service 中分离。便于单元测试使用。
        2011-08-27  + 新增 service_async 函数。
        2011-08-28  * 重构 Pack 为一个独立类。
        2011-08-29  * 取消 Timeout。
"""
import time
import copy
import datetime
import json
import traceback
import pkgutil
import time, sched
import threading

from sys import stderr, argv
from multiprocessing import cpu_count, Process


try:
    from os import wait, fork, getpid, getppid, killpg, waitpid
    from signal import signal, pause, SIGCHLD, SIGINT, SIGTERM, SIGUSR1, SIGUSR2, SIG_IGN
    iswin = False
except:
    iswin = True
    print('some function only support unix and linux ')

try:
    from redis import Redis, ConnectionError
except:
    print("Redis is no exist!")

from json import loads, dumps

from kpages.log import log
from kpages.context import LogicContext, get_context
from kpages.utility import get_members


def staticclass(cls):
    def new(cls, *args, **kwargs):
        raise RuntimeError("Static Class")

    setattr(cls, "__new__", staticmethod(new))
    return cls


def srvcmd(cmd, sub_mode=0):
    """
    sub_mode : 0 时所有消费端都可以执行，-1时，只允许其中一个消费端执行
    """
    def actual(func):
        func.__service__ = cmd
        func.__sub_mode__ = sub_mode
        return func

    return actual


def srvtimer(delay):
    def actual(func):
        func.__timer__ = delay
        return func

    return actual

class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        
        return json.JSONEncoder.default(self, obj)
            

@staticclass
class Pack(object):

    @staticmethod
    def unpack(data):
        try:
            pack = loads(data)
            return pack["cmd"], pack["data"]
        except:
            return 'undefined cmd', None

    @classmethod
    def async_send_pack(cls, cmd, data, channel=None):
        r = get_context().get_redis()
        pack = dumps(dict(cmd=cmd, data=data), cls=DateTimeEncoder)
        r.publish(channel or __conf__.SERVICE_CHANNEL, pack)

    @classmethod
    def async_queue(cls, cmd, data, channel=None):
        r = get_context().get_redis()
        pack = dumps(dict(cmd=cmd, data=data), cls=DateTimeEncoder)
        r.rpush(channel or __conf__.SERVICE_CHANNEL, pack)


class PSConsumer(object):
    def __init__(self, channel, host="localhost"):
        self._channel = channel
        self._host = host

    def subscribe(self):
        h, p = self._host.split(":") if ":" in self._host else (self._host, 6379)
        self._redis = Redis(host=h, port = p)
        self._pubsub = self._redis.pubsub()
        self._pubsub.subscribe(self._channel)

    def consume(self):
        try:
            data = self._pubsub.listen().next()["data"]
            return Pack.unpack(data)
        except Exception as e:
            self._pubsub.subscribe(self._channel)
            return None, None


    def close(self):
        if hasattr(self, "_redis"):
            self._redis.connection_pool.disconnect()


class Service(object):
    """
        MQ Service

        Demo:
            ./service.py [host] [channel]

            host    default "localhost"
            channel defailt  settings.py SERVICE_CHANNEL
    """
    def __init__(self, host=None, channel=None, callback=None):
        self._host = host or "localhost"
        self._channel = channel or __conf__.SERVICE_CHANNEL
        self._processes = (__conf__.DEBUG and 1 or cpu_count()) * __conf__.CPU_MULTIPLE

        self._ps_consumer = PSConsumer(self._channel, self._host)
        self._services = self._get_services()

        if callback:
            kwargs = dict(
                host=self._host,
                channel=self._channel,
                processes=self._processes,
                services=self._services,
                debug=__conf__.DEBUG
            )
            callback(**kwargs)

    def _signal(self):

        self._parent = getpid()
        def sig_handler(signum, frame):
            pid = getpid()

            if signum in (SIGINT, SIGTERM):
                if pid == self._parent:
                    signal(SIGCHLD, SIG_IGN)
                    killpg(self._parent, SIGUSR1)
            elif signum == SIGCHLD:
                if pid == self._parent:
                    print >> stderr, "sub process {0} exit...".format(wait())
                    killpg(self._parent, SIGUSR1)
            elif signum == SIGUSR1:
                print("process {0} exit...".format(pid == self._parent and pid or (pid, self._parent)))
                exit(0)

        signal(SIGINT, sig_handler)
        signal(SIGTERM, sig_handler)
        signal(SIGCHLD, sig_handler)
        signal(SIGUSR1, sig_handler)

    def _get_services(self):
        try:
            members = get_members(
                __conf__.JOB_DIR, lambda o: hasattr(o, "__service__"))
            svrs = {}
            for k,v in members.items():
                if not svrs.get(v.__service__, None):
                    svrs[v.__service__] = []

                svrs[v.__service__].append(v)
            return svrs

        except Exception as e:
            traceback.print_exc()
            return {}


    def _queue_consumer(self):
        for i in range(self._processes):
            if not iswin:
                if fork() > 0:
                    continue

            with LogicContext():
                log_queue_consumer = log("log/service-queue-consumer", level = 'info' if not __conf__.DEBUG else 'debug')
                channel = ['REDIS_QUEUE_1', self._channel]
                r = get_context().get_redis()
                while True:
                    try:
                        channel, data = r.brpop(channel)
                        cmd, data = Pack.unpack(data)
                        srv_funcs = self._services.get(cmd, ())

                        ps = []
                        for func in srv_funcs:
                            try:
                                func(data)
                            except Exception as e:
                                log_queue_consumer.error("{}".format(traceback.format_exc()))

                    except (SystemExit, KeyboardInterrupt) as e:
                        log_queue_consumer.error("{}".format(traceback.format_exc()))
                        break;
                    except Exception as e:
                        log_queue_consumer.error("{}".format(traceback.format_exc()))
                        time.sleep(1)
                        continue

            exit(0)

    def ps_consumer(self):
        if not iswin:
            if fork() > 0:
                return

        with LogicContext():
            self._ps_consumer.subscribe()
            log_psconsumer = log("log/service-psconsumer", level = 'info' if not __conf__.DEBUG else 'debug')
            while True:
                try:
                    cmd, data = self._ps_consumer.consume()
                    srv_funcs = self._services.get(cmd, ())

                    ps = []
                    for func in srv_funcs:
                        try:
                            func(data)
                        except Exception as e:
                            log_psconsumer.error("{}".format(traceback.format_exc()))

                except (SystemExit, KeyboardInterrupt) as e:
                    log_psconsumer.error("{}".format(traceback.format_exc()))
                    break;
                except Exception as e:
                    log_psconsumer.error("{}".format(traceback.format_exc()))
                    time.sleep(1)
                    continue

            exit(0)

    def run(self):
        if not iswin:
            self._signal()

        try:
            self.ps_consumer()
            self._queue_consumer()
        except RuntimeError:
            print("Is running?")
            exit(-1)

        #添加代码改动监控
        while True:
            pause()


service_async = Pack.async_send_pack
service_async_queue = Pack.async_queue

__all__ = ["Service", "srvcmd", "service_async", "srvtimer", "service_async_queue"]




#coding=utf-8

import sys
import logging
import logging.handlers

#日志内容最大值
MAX_BYTES = 1024*1024*100

#最大日志保留数
B_COUNT = 30

#日志等级
LEVELS = {'debug': logging.DEBUG,
          'info': logging.INFO,
          'warning': logging.WARNING,
          'error': logging.ERROR,
          'critical': logging.CRITICAL}

class log:
    """
    封装logging类
    >>level 可以输出的最低级别
    >>type    0:输出到前台和文件 1:输出到文件
    """
    def __init__(self, name, level = 'debug', type = 1):
        debug_file = name+"-debug.log"
        error_file = name+"-error.log"

        self.debug_logger = logging.getLogger(debug_file)
        self.error_logger = logging.getLogger(error_file)

        level = LEVELS.get(level, logging.NOTSET)
        self.debug_logger.setLevel(level)

        dfh = logging.handlers.TimedRotatingFileHandler(debug_file, when = 'midnight', backupCount = B_COUNT)
        efh = logging.handlers.RotatingFileHandler(error_file, maxBytes=MAX_BYTES, backupCount = B_COUNT)

        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

        dfh.setFormatter(formatter)
        efh.setFormatter(formatter)

        self.debug_logger.addHandler(dfh)
        self.error_logger.addHandler(efh)

        if type == 0:
            ch = logging.StreamHandler()
            ch.setFormatter(formatter)
            self.debug_logger.addHandler(ch)
            self.error_logger.addHandler(ch)

    def debug(self, message):
        """调试信息"""
        self.debug_logger.debug(message)

    def info(self, message):
        """普通消息"""
        self.debug_logger.info(message)

    def error(self, message):
        """错误消息"""
        self.error_logger.error(message)


def test():
    argn = len(sys.argv)
    if argn > 2:
        mylog = log("test", sys.argv[1], sys.argv[2])
    elif argn > 1:
        mylog = log("test", sys.argv[1])
    else:
        mylog = log("test")
              
    print("Log testing ...")
    mylog.debug('This is a debug message')
    mylog.info('This is an info message')
    mylog.error('This is an error message')
        

if __name__ == '__main__':
    test()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
    author comger@gmail.com
"""
import sys
import json

from datetime import datetime
from pprint import pprint
from kpages import run

def callback(app):

    print("Start time: {0}".format(datetime.now().isoformat(" ")))
    print("Config Params")
    for k in sorted(app.settings.keys()):
        if k.startswith("__"):
            continue
        print("  {0:<40} : {1}".format(k, app.settings[k]))

    
    print("UI Modules")
    for k,v in app.uimodules.items():
        print(' {0:<20} :{1}'.format(k,v.__doc__))


    print("UI  Methods")
    for k,v in app.uimethods.items():
        print(' {0:<20} :{1}'.format(k,v.__intro__))

    
    print("Router Handlers")
    for h in app.handlers:
        print('  {0:<50} : {1}'.format(h[1],h[0]))

if __name__ == "__main__":
    try:
        run(callback)
    except KeyboardInterrupt:
        print('exit server ')

# vim: ts=4 sw=4 sts=4 expandtab



# -*- coding:utf-8 -*-
from kpages.router import *
from kpages.context import *
from kpages.utility import *
from kpages.app import *
from kpages.utest import *
from kpages.service import *
from kpages.api2doc import *



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
    author comger@gmail.com
"""
import tornado.ioloop

from optparse import OptionParser
from kpages import LogicContext, run_test, pro_test, run_doc, reflesh_config, set_default_encoding


def _get_opt():
    parser = OptionParser("%prog [options]", version="%prog v0.9")
    parser.add_option("--config", dest="config",
                      default='setting.py', help="config for app")
    parser.add_option("--test", dest="test", default=None, help="utest module")
    parser.add_option("--doc", dest="doc", default=None, help="doc all router api to markdown")
    parser.add_option(
        "--pro", dest="pro", default=None, help="profile for method")
    return parser.parse_args()

if __name__ == "__main__":
    try:
        set_default_encoding()
        opts, args = _get_opt()
        reflesh_config(opts.config)

        if opts.test is not None:
            m = opts.test
            if m == 'all':
                m = None
            with LogicContext():
                run_test(m)

        elif opts.pro is not None:
            with LogicContext():
                pro_test(opts.pro)

        elif opts.doc is not None:
            with LogicContext():
                run_doc(opts.doc)

        tornado.ioloop.IOLoop.instance().start()
    except KeyboardInterrupt:
        print('exit tool ')

# vim: ts=4 sw=4 sts=4 expandtab
